import logging
from enum import Enum
from typing import Optional
from datetime import datetime, timedelta
from whendo.core.util import Now
from whendo.core.hooks import DispatcherHooks
from whendo.core.action import Action


logger = logging.getLogger(__name__)


class DispActionMode(str, Enum):
    """
    usage:
        DispActionMode.field
    """

    field = "field"
    data = "data"


class DispatcherAction(Action):
    """
    Instances of subclasses of DispatcherAction invoke Dispatcher
    methods directly by way of the class, DispatcherHooks.

    The class's fields determine the variable set necessary for
    successful execution of the action. With this in mind, the
    action draws these variable values from two sources: the instance's
    fields and the 'data' argument of the execute method.

    The mode field prioritizes which source to favor for the execution
    of the action. Modes are either "field" or "data".

    field:

        the instance variables are the first choice for values used
        in the action's execute method. If an instance variable is
        absent (==None), then the execute method looks for the value
        in the incoming 'data' dictionary.

        'field' is the default mode.

    data:

        the dictionary items in the incoming 'data' dictionary argument
        are the first choice for values used in the execute method. If
        a dictionary key is absent, then the execute method looks for the
        value in the instance variables of the action object.

    [Note the use of self.getResult()]
    """

    dispatcher_action: str = "dispatcher_action"
    mode: DispActionMode = DispActionMode.field

    def compute_args(self, args: dict, data: dict = None):
        """
        Implements field vs. data mode arg construction logic.
        """
        if data:
            data_result = self.get_result(data)
            if isinstance(data_result, dict):
                data_args = {
                    arg: data_result[arg] for arg in args.keys() if arg in data_result
                }
                if self.mode == DispActionMode.data:
                    args.update(data_args)
                else:
                    data_args.update(args)
                    args = data_args
        return args


class ScheduleProgram(DispatcherAction):
    schedule_program: str = "schedule_program"
    program_name: Optional[str] = None
    start: Optional[datetime] = None
    stop: Optional[datetime] = None

    def description(self):
        return f"This action unschedules a program with mode ({self.mode}) and fields: program_name ({self.program_name}), start ({self.start}), stop ({self.stop})."

    def execute(self, tag: str = None, data: dict = None):
        args = {
            "program_name": self.program_name,
            "start": self.start,
            "stop": self.stop,
        }
        args = self.compute_args(args, data)
        DispatcherHooks.schedule_program(**args)
        result = f"program ({args['program_name']}) scheduled, start({args['start']}) stop({args['stop']})"
        return self.action_result(result=result, data=data, extra=args)


class UnscheduleProgram(DispatcherAction):
    unschedule_program: str = "unschedule_program"
    program_name: Optional[str] = None

    def description(self):
        return f"This action unschedules a program with mode ({self.mode}) and field: program_name ({self.program_name})."

    def execute(self, tag: str = None, data: dict = None):
        args = {
            "program_name": self.program_name,
        }
        args = self.compute_args(args, data)
        DispatcherHooks.unschedule_program(**args)
        result = f"program ({args['program_name']}) unscheduled"
        return self.action_result(result=result, data=data, extra=args)


class ScheduleAction(DispatcherAction):
    schedule_action: str = "schedule_action"
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None

    def description(self):
        return f"This action schedules an action with mode ({self.mode}) and fields: scheduler_name ({self.scheduler_name}), action_name ({self.action_name})."

    def execute(self, tag: str = None, data: dict = None):
        # gather all of the args that can participate
        args = {"scheduler_name": self.scheduler_name, "action_name": self.action_name}
        args = self.compute_args(args, data)
        DispatcherHooks.schedule_action(**args)
        result = f"action ({args['action_name']}) scheduled using scheduler ({args['scheduler_name']})"
        return self.action_result(result=result, data=data, extra=args)


class UnscheduleSchedulerAction(DispatcherAction):
    unschedule_scheduler_action: str = "unschedule_scheduler_action"
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None

    def description(self):
        return f"This action unschedules a scheduler/action with mode ({self.mode}) and fields: scheduler_name ({self.scheduler_name}), action_name ({self.action_name})."

    def execute(self, tag: str = None, data: dict = None):
        # gather all of the args that can participate
        args = {"scheduler_name": self.scheduler_name, "action_name": self.action_name}
        args = self.compute_args(args, data)
        DispatcherHooks.unschedule_scheduler_action(**args)
        result = f"action ({args['action_name']}) unscheduled from scheduler ({args['scheduler_name']})"
        return self.action_result(result=result, data=data, extra=args)


class UnscheduleScheduler(DispatcherAction):
    unschedule_scheduler: str = "unschedule_scheduler"
    scheduler_name: Optional[str] = None

    def description(self):
        return f"This action unschedules a scheduler with mode ({self.mode}) and field: scheduler_name ({self.scheduler_name})."

    def execute(self, tag: str = None, data: dict = None):
        # gather all of the args that can participate
        args = {"scheduler_name": self.scheduler_name}
        args = self.compute_args(args, data)
        DispatcherHooks.unschedule_scheduler(**args)
        result = f"scheduler ({args['scheduler_name']}) unscheduled"
        return self.action_result(result=result, data=data, extra=args)


class DeferAction(DispatcherAction):
    defer_action: str = "defer_action"
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None
    wait_until: Optional[datetime] = None

    def description(self):
        return f"This action defers a schedule/action with mode ({self.mode}) and fields: scheduler_name ({self.scheduler_name}), action_name ({self.action_name}), wait_until ({self.wait_until})."

    def execute(self, tag: str = None, data: dict = None):
        # gather all of the args that can participate
        args = {
            "scheduler_name": self.scheduler_name,
            "action_name": self.action_name,
            "wait_until": self.wait_until,
        }
        args = self.compute_args(args, data)
        DispatcherHooks.defer_action(**args)
        result = f"action ({args['action_name']}) using scheduler ({args['scheduler_name']}) deferred until ({args['wait_until']})"

        return self.action_result(result=result, data=data, extra=args)


class ExpireAction(DispatcherAction):
    expire_action: str = "expire_action"
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None
    expire_on: Optional[datetime] = None

    def description(self):
        return f"This action expires a schedule/action with mode ({self.mode}) and fields: scheduler_name ({self.scheduler_name}), action_name ({self.action_name}), expire_on ({self.expire_on})."

    def execute(self, tag: str = None, data: dict = None):
        # gather all of the args that can participate
        args = {
            "scheduler_name": self.scheduler_name,
            "action_name": self.action_name,
            "expire_on": self.expire_on,
        }
        args = self.compute_args(args, data)
        DispatcherHooks.expire_action(**args)
        result = f"action ({args['action_name']}) using scheduler ({args['scheduler_name']}) expiring on ({args['expire_on']})"
        return self.action_result(result=result, data=data, extra=args)


class ClearAllDeferredActions(DispatcherAction):
    clear_all_deferred_actions: str = "clear_all_deferred_actions"

    def description(self):
        return "Removes all deferred scheduled actions."

    def execute(self, tag: str = None, data: dict = None):
        DispatcherHooks.clear_all_deferred_actions()
        result = "All deferred scheduled actions removed."
        return self.action_result(result=result, data=data)


class ClearAllExpiringActions(DispatcherAction):
    clear_all_expiring_actions: str = "clear_all_expiring_actions"

    def description(self):
        return "Removes all expiring scheduled actions."

    def execute(self, tag: str = None, data: dict = None):
        DispatcherHooks.clear_all_expiring_actions()
        result = "All expiring scheduled actions removed."
        return self.action_result(result=result, data=data)



from pydantic import BaseModel
from typing import Dict, List, Set, Callable
from datetime import datetime
from collections import namedtuple
import logging
from .util import Now, str_to_dt, dt_to_str


logger = logging.getLogger(__name__)


class SchedulerActions(BaseModel):
    """
    Objects of this class provide the fundamental mapping of schedulers to actions.

    These objects also form the basis for the extended properties of schedulers
    such as the deferral of starting a scheduler and the expiration of a running
    scheduler. See the Dispatcher class for usage.
    """
    scheduler_name: str
    action_names: Set[str] = set()


class ScheduledActions(BaseModel):
    scheduler_actions: List[SchedulerActions] = []

    def add(self, scheduler_name: str, action_name: str):
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                sa.action_names.add(action_name)
                return
        new_scheduler_action = SchedulerActions(
            scheduler_name=scheduler_name, action_names={action_name}
        )
        self.scheduler_actions.append(new_scheduler_action)

    def actions(self, scheduler_name: str):
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                return sa.action_names
        return set()

    def delete(self, scheduler_name: str, action_name: str):
        to_remove = None
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                if action_name in sa.action_names:
                    sa.action_names.remove(action_name)
                    if len(sa.action_names) == 0:
                        to_remove = sa
                break
        if to_remove:
            self.scheduler_actions.remove(to_remove)
            return to_remove.scheduler_name
        else:
            return None

    def delete_action(self, action_name: str):
        removed_scheduler_names: Set[str] = set()
        for scheduler_name in self.scheduler_names():
            to_remove = self.delete(scheduler_name, action_name)
            if to_remove:
                removed_scheduler_names.add(to_remove)
        return removed_scheduler_names

    def delete_scheduler(self, scheduler_name: str):
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                self.scheduler_actions.remove(sa)
                return scheduler_name
        return None

    def scheduler_names(self):
        return set(sa.scheduler_name for sa in self.scheduler_actions)

    def action_names(self):
        action_names = set()
        for sa in self.scheduler_actions:
            action_names.update(sa.action_names)
        return action_names

    def action_count(self):
        return len(self.action_names())

    def clear(self):
        self.scheduler_actions.clear()


DeferredProgram = namedtuple("DeferredProgram", ["program_name", "start", "stop"])


class DeferredPrograms(BaseModel):
    deferred_programs: Set[DeferredProgram] = set()

    def add(self, deferred_program: DeferredProgram):
        self.deferred_programs.add(deferred_program)

    def pop(self):
        filtered = set(dp for dp in self.deferred_programs if dp.start < Now.dt())
        self.deferred_programs = self.deferred_programs - filtered
        return filtered

    def clear_program(self, program_name: str):
        self.deferred_programs = set(
            dp for dp in self.deferred_programs if dp.program_name != program_name
        )

    def clear(self):
        self.deferred_programs = set()

    def count(self):
        return len(self.deferred_programs)


class DatedScheduledActions(BaseModel):
    dated_scheduled_actions: Dict[str, ScheduledActions] = {}

    def check_for_dated_actions(
        self,
        schedule_update_thunk: Callable,
        verb: str,
    ):
        now = Now.dt()
        to_remove = []
        for date_time_str in self.dated_scheduled_actions:
            date_time = str_to_dt(date_time_str)
            if date_time < now:
                scheduled_actions = self.dated_scheduled_actions[date_time_str]
                for scheduler_name in scheduled_actions.scheduler_names():
                    for action_name in scheduled_actions.actions(scheduler_name):
                        try:
                            schedule_update_thunk(scheduler_name, action_name)
                        except Exception as exception:
                            logger.error(
                                f"failed to {verb} action ({action_name}) under ({scheduler_name}) as of ({date_time_str})",
                                exception,
                            )
                to_remove.append(date_time_str)
        for date_time_str in to_remove:  # modify outside the previous for-loop
            self.dated_scheduled_actions.pop(date_time_str)

    def apply_date(
        self,
        scheduler_name: str,
        action_name: str,
        date_time: datetime,
    ):
        date_time_str = dt_to_str(date_time)
        if date_time_str not in self.dated_scheduled_actions:
            self.dated_scheduled_actions[date_time_str] = ScheduledActions()
        scheduled_actions = self.dated_scheduled_actions[date_time_str]
        scheduled_actions.add(scheduler_name, action_name)

    def action_count(self):
        return sum(sa.action_count() for sa in self.dated_scheduled_actions.values())

    def delete_dated_action(self, action_name: str):
        date_time_str_to_remove = []
        for date_time_str in self.dated_scheduled_actions:
            scheduled_actions = self.dated_scheduled_actions[date_time_str]
            scheduled_actions.delete_action(action_name)
            if len(scheduled_actions.scheduler_names()) == 0:
                date_time_str_to_remove.append(date_time_str)
        for date_time_str in date_time_str_to_remove:
            self.dated_scheduled_actions.pop(date_time_str)

    def delete_dated_scheduler(self, scheduler_name: str):
        date_time_str_to_remove = []
        for date_time_str in self.dated_scheduled_actions:
            scheduled_actions = self.dated_scheduled_actions[date_time_str]
            scheduled_actions.delete_scheduler(scheduler_name)
            if len(scheduled_actions.scheduler_names()) == 0:
                date_time_str_to_remove.append(date_time_str)
        for date_time_str in date_time_str_to_remove:
            self.dated_scheduled_actions.pop(date_time_str)

    def clear(self):
        self.dated_scheduled_actions.clear()



"""
These functions resolve dispatcher element instances from supplied dictionaries.
"""
import logging
from pydantic import ValidationError
from .action import Action
from .scheduler import Scheduler
from .program import Program
from .util import FilePathe, resolve_instance


logger = logging.getLogger(__name__)


def resolve_action(dictionary: dict, check_for_found_class: bool = True):
    return resolve_instance(
        Action, dictionary, check_for_found_class=check_for_found_class
    )


def resolve_scheduler(dictionary: dict, check_for_found_class: bool = True):
    return resolve_instance(
        Scheduler, dictionary, check_for_found_class=check_for_found_class
    )


def resolve_program(dictionary: dict, check_for_found_class: bool = True):
    return resolve_instance(
        Program, dictionary, check_for_found_class=check_for_found_class
    )


def resolve_file_pathe(dictionary: dict, check_for_found_class: bool = True):
    try:
        return FilePathe(**dictionary)
    except ValidationError as error:
        if check_for_found_class:
            raise error
        else:
            return dictionary



"""
This script establishes the top FastAPI path and all sub-path routers.
It also initializes the Dispatcher for the calling ASGI server.
"""
from fastapi import FastAPI
from whendo.core.dispatcher import DispatcherSingleton
from whendo.api.router import actions, schedulers, dispatcher, jobs, execution, programs
from whendo.api.shared import set_dispatcher
import whendo.log.init

app = FastAPI()


@app.get("/")
async def root():
    return "whengo API server started"


dispatcher_instance = DispatcherSingleton.get()

app.include_router(set_dispatcher(actions.router, dispatcher_instance))
app.include_router(set_dispatcher(schedulers.router, dispatcher_instance))
app.include_router(set_dispatcher(dispatcher.router, dispatcher_instance))
app.include_router(set_dispatcher(jobs.router, dispatcher_instance))
app.include_router(set_dispatcher(execution.router, dispatcher_instance))
app.include_router(set_dispatcher(programs.router, dispatcher_instance))



from datetime import datetime
from typing import Callable


class DispatcherHooks:
    schedule_program_thunk: Callable
    unschedule_program_thunk: Callable
    schedule_action_thunk: Callable
    unschedule_scheduler_action_thunk: Callable
    unschedule_scheduler_thunk: Callable
    defer_action_thunk: Callable
    expire_action_thunk: Callable
    clear_all_deferred_actions_thunk: Callable
    clear_all_expiring_actions_thunk: Callable

    @classmethod
    def init(
        cls,
        schedule_program_thunk: Callable,
        unschedule_program_thunk: Callable,
        schedule_action_thunk: Callable,
        unschedule_scheduler_action_thunk: Callable,
        unschedule_scheduler_thunk: Callable,
        defer_action_thunk: Callable,
        expire_action_thunk: Callable,
        clear_all_deferred_actions_thunk: Callable,
        clear_all_expiring_actions_thunk: Callable,
    ):
        cls.schedule_program_thunk = schedule_program_thunk
        cls.unschedule_program_thunk = unschedule_program_thunk
        cls.schedule_action_thunk = schedule_action_thunk
        cls.unschedule_scheduler_action_thunk = unschedule_scheduler_action_thunk
        cls.unschedule_scheduler_thunk = unschedule_scheduler_thunk
        cls.defer_action_thunk = defer_action_thunk
        cls.expire_action_thunk = expire_action_thunk
        cls.clear_all_deferred_actions_thunk = clear_all_deferred_actions_thunk
        cls.clear_all_expiring_actions_thunk = clear_all_expiring_actions_thunk

    @classmethod
    def schedule_program(cls, program_name: str, start: datetime, stop: datetime):
        return cls.schedule_program_thunk(
            program_name=program_name, start=start, stop=stop
        )

    @classmethod
    def unschedule_program(cls, program_name: str):
        return cls.unschedule_program_thunk(program_name=program_name)

    @classmethod
    def schedule_action(cls, scheduler_name: str, action_name: str):
        return cls.schedule_action_thunk(
            scheduler_name=scheduler_name, action_name=action_name
        )

    @classmethod
    def unschedule_scheduler_action(cls, scheduler_name: str, action_name: str):
        return cls.unschedule_scheduler_action_thunk(
            scheduler_name=scheduler_name, action_name=action_name
        )

    @classmethod
    def unschedule_scheduler(cls, scheduler_name: str):
        return cls.unschedule_scheduler_thunk(scheduler_name=scheduler_name)

    @classmethod
    def defer_action(cls, scheduler_name: str, action_name: str, wait_until: datetime):
        return cls.defer_action_thunk(
            scheduler_name=scheduler_name,
            action_name=action_name,
            wait_until=wait_until,
        )

    @classmethod
    def expire_action(cls, scheduler_name: str, action_name: str, expire_on: datetime):
        return cls.expire_action_thunk(
            scheduler_name=scheduler_name, action_name=action_name, expire_on=expire_on
        )

    @classmethod
    def clear_all_deferred_actions(cls):
        return cls.clear_all_deferred_actions_thunk()

    @classmethod
    def clear_all_expiring_actions(cls):
        return cls.clear_all_expiring_actions_thunk()



from pydantic import BaseModel
from typing import Dict, Any, Optional
import logging
from collections import namedtuple
from .util import object_info, SystemInfo, Now


logger = logging.getLogger(__name__)


class Action(BaseModel):
    """
    Actions get something done.
    """

    def description(self):
        return "This has no description."

    def execute(self, tag: str = None, data: dict = None):
        """
        This method typically attempts to do something useful and return something useful
        """
        return {"result": "not sure what happened"}

    def info(self):
        return object_info(self)

    def flat(self):
        return self.json()

    def local_host(self):
        return SystemInfo.get()["host"]

    def local_port(self):
        return SystemInfo.get()["port"]

    def local_time(self):
        return SystemInfo.get()["current"]

    def local_info(self):
        return {
            "host": self.local_host(),
            "port": self.local_port(),
            "time": self.local_time(),
        }

    def action_result(self, result: Any = None, data: dict = None, extra: dict = None):
        output = {}
        if result is not None:
            output["result"] = result
        if data is not None:
            output["data"] = data
        if extra is not None:
            output["extra"] = extra
        output["info"] = self.info()
        return output

    def get_result(self, x):
        if isinstance(x, dict):
            if "result" in x:
                return x["result"]
        return x



from typing import List, Dict, Any, Optional
from enum import Enum
import json
import logging
from collections import namedtuple
from whendo.core.action import Action
from whendo.core.exception import TerminateSchedulerException


logger = logging.getLogger(__name__)


def processing_results(
    result: Any,
    processing_count: int = 0,
    success_count: int = 0,
    exception_count: int = 0,
    successful_actions: List[Dict[Any, Any]] = [],
    exception_actions: List[Dict[Any, Any]] = [],
):
    return {
        "result": result,
        "processing_count": processing_count,
        "success_count": success_count,
        "exception_count": exception_count,
        "successful_actions": successful_actions,
        "exception_actions": exception_actions,
    }


class Arg(Action):
    arg: str = "arg"
    data: dict

    def description(self):
        return f"This action returns ({self.payload})."

    def execute(self, tag: str = None, data: dict = None):
        return self.action_result(result=self.get_result(self.data), data=data)


class RaiseIfEqual(Action):
    """
    If fails to raise, pass through the supplied the result value
    from the supplied execute argument, data.
    """

    raise_if_equal: str = "raise_if_equal"
    value: Any

    def description(self):
        return f"This action raises an exception if the result in the supplied argument, data, is equal to ({self.value}). Otherwise it returns the value untouched."

    def execute(self, tag: str = None, data: dict = None):
        value = self.value
        data_value = self.get_result(data)
        should_raise = data_value == value
        if should_raise:
            raise ValueError(
                f"exception thrown because ({value}) == ({data_value}); info ({self.info()})"
            )
        else:
            return self.action_result(result=data_value, data=data)


class Terminate(Action):
    terminate: str = "terminate"

    def description(self):
        return f"This action raises a TerminateScheduler exception."

    def execute(self, tag: str = None, data: dict = None):
        raise TerminateSchedulerException(value=tag)


class Failure(Action):
    """ acts like False """

    failure: str = "failure"

    def description(self):
        return f"This action always fails. It serves a role similar to the bool, False."

    def execute(self, tag: str = None, data: dict = None):
        raise Exception("purposely unsuccessful execution", self.json())


class Success(Action):
    """ acts like True. It also serves as a fixed point execute implementation. """

    success: str = "success"

    def description(self):
        return (
            f"This action always succeeds. It serves a role similar to the bool, True."
        )

    def execute(self, tag: str = None, data: dict = None):
        return self.action_result(result=self.get_result(data), data=data)


class Fail(Action):
    """ acts like negation """

    fail: str = "fail"
    operand: Action

    def description(self):
        return f"If the operand ({self.operand}) fails, Not succeeds, otherwise fails (throwing an Exception). It serves a role similar to logical negation."

    def execute(self, tag: str = None, data: dict = None):
        operand_result = None
        try:
            operand_result = self.operand.execute(tag=tag, data=data)
        except Exception as exception:
            operand_result = exception
        if isinstance(operand_result, Exception):
            extra = {"execption": str(operand_result)}
            return self.action_result(result=True, data=data, extra=extra)
        else:
            raise Exception(
                "exception generated; action execution treated as a failure",
                self.json(),
            )


class ListOpMode(str, Enum):
    ALL = "all"
    UNTIL_SUCCESS = "until_success"
    UNTIL_FAILURE = "until_failure"


class ListAction(Action):
    """
    executes actions based on:

        ListOpMode.ALL:   executes all, irrespective of individual outcomes
        ListOpMode.UNTIL_SUCCESS:    executes until the first successful action (no exception)
        ListOpMode.UNTIL_FAILURE:   executes until the first exception (failure)

    Intended to be abstract class; not intended to be instantiated. Its subclasses,
    All, Or, and And, should be used instead.
    """

    op_mode: ListOpMode
    actions: List[Action] = []
    exception_on_no_success: bool = False
    include_processing_info: bool = False

    def execute(self, tag: str = None, data: dict = None):
        processing_info = process_actions(
            data=data,
            op_mode=self.op_mode,
            tag=tag,
            actions=self.actions,
            successful_actions=[],
            exception_actions=[],
        )
        if (
            processing_info["processing_count"] > 0
            and processing_info["success_count"] == 0
            and self.exception_on_no_success
        ):
            raise Exception(
                "exception generated; all actions failing treated as a list action failure",
                json.dumps(self.dict()),
                json.dumps(processing_info),
            )
        extra = processing_info if self.include_processing_info else None
        return self.action_result(
            result=self.get_result(processing_info["result"]), data=data, extra=extra
        )


class All(ListAction):
    """
    executes all actions
    """

    _all: str = "all"
    op_mode: ListOpMode = ListOpMode.ALL

    def description(self):
        return f"This action executes all of these actions in order: ({self.actions})."

    def execute(self, tag: str = None, data: dict = None):
        return super().execute(tag=tag, data=data)


class UntilSuccess(ListAction):
    """
    executes actions until first success
    """

    until_success: str = "until_success"
    op_mode: ListOpMode = ListOpMode.UNTIL_SUCCESS

    def description(self):
        return f"This action executes all of these actions in order until the first success: ({self.actions}). It serves a role similar to logical or."

    def execute(self, tag: str = None, data: dict = None):
        return super().execute(tag=tag, data=data)


class UntilFailure(ListAction):
    """
    executes actions until first failure
    """

    until_failure: str = "until_failure"
    op_mode: ListOpMode = ListOpMode.UNTIL_FAILURE

    def description(self):
        return f"This action executes all of these actions in order until the first failure: ({self.actions}). It serves a role similar to logical and."

    def execute(self, tag: str = None, data: dict = None):
        return super().execute(tag=tag, data=data)


def process_actions(
    op_mode: ListOpMode,
    actions: List[Action],
    tag: str = None,
    data: Dict[Any, Any] = None,
    successful_actions: List[Dict[Any, Any]] = [],
    exception_actions: List[Dict[Any, Any]] = [],
    processing_count: int = 0,
    success_count: int = 0,
    exception_count: int = 0,
):
    """
    services the three action list classes, ALL, UNTIL_SUCCESS, and UNTIL_FAILURE
    """
    exceptions: List[Exception] = []
    loop_data = data
    for action in actions:
        try:
            result = action.execute(tag=tag, data=loop_data)
            logger.info(
                f"ListAction: tag ({tag}); executed action ({action}); loop data ({loop_data})"
            )
        except Exception as exception:
            result = exception
            logger.exception(
                f"ListAction: tag ({tag}); error while executing action ({action}); loop data ({loop_data})",
                exc_info=exception,
            )
        processing_count += 1
        if isinstance(result, Exception):
            exception_count += 1
            exception_dict = action.dict().copy()
            exception_dict.update({"exception": str(result)})
            exception_actions.append(exception_dict)
            if isinstance(result, TerminateSchedulerException):
                result.value = processing_results(
                    loop_data,
                    processing_count,
                    success_count,
                    exception_count,
                    successful_actions,
                    exception_actions,
                )
                raise result
            if op_mode == ListOpMode.UNTIL_FAILURE:  # stop after first failure
                break
            loop_data = exception_dict
        else:
            success_count += 1
            successful_actions.append(action.dict())
            loop_data = result
            if op_mode == ListOpMode.UNTIL_SUCCESS:  # stop after first success
                break
    return processing_results(
        loop_data,
        processing_count,
        success_count,
        exception_count,
        successful_actions,
        exception_actions,
    )


class IfElse(Action):
    """
    executes actions based on:

        if test_action is successful, executes if_action
        otherwise executes else_action

    """

    if_else: str = "if_else"
    test_action: Action
    else_action: Action
    if_action: Optional[Action]
    exception_on_no_success: bool = False
    include_processing_info: bool = False

    def logger_info(self, s: str):
        logger.info(s)
        print(s)

    def logger_exception(self, s: str, exc_info: object):
        logger.info(s, exc_info)
        print(s)

    def description(self):
        return f"If  action ({self.test_action}) succeeds, then IfElse executes ({self.if_action}), otherwise executes ({self.else_action})"

    def execute(self, tag: str = None, data: dict = None):
        print(f"IfElse::execute::data=({data})")
        try:
            test_result = self.test_action.execute(tag=tag, data=data)
            self.logger_info(
                f"IfElse: tag ({tag}); successfully executed 'test' action ({self.test_action})"
            )
        except Exception as exception:
            test_result = exception
            self.logger_exception(
                f"IfElse: tag ({tag}); error while executing 'test' action ({self.test_action}) exception ({exception})",
                exc_info=exception,
            )
        processing_count = 1
        success_count = 0
        exception_count = 0
        successful_actions: List[Dict[Any, Any]] = []
        exception_actions: List[Dict[Any, Any]] = []
        else_test = isinstance(test_result, Exception)
        if else_test:  # execute the else action
            exception_count = 1
            exception_dict = self.test_action.dict().copy()
            exception_dict.update({"exception": str(test_result)})
            exception_actions.append(exception_dict)
            if isinstance(test_result, TerminateSchedulerException):
                test_result.value = processing_results(
                    data,
                    processing_count,
                    success_count,
                    exception_count,
                    successful_actions,
                    exception_actions,
                )
                raise test_result

            try:
                else_result = self.else_action.execute(tag=tag, data=data)
                self.logger_info(
                    f"IfElse: tag ({tag}); successfully executed 'else' action ({self.else_action})"
                )
            except Exception as exception:
                else_result = exception
                self.logger_exception(
                    f"IfElse: tag ({tag}); error while executing 'else' action ({self.else_action}) exception ({exception})",
                    exc_info=exception,
                )
            processing_count += 1
            if isinstance(else_result, Exception):
                exception_count += 1
                exception_dict = self.else_action.dict().copy()
                exception_dict.update({"exception": str(else_result)})
                exception_actions.append(exception_dict)
                if isinstance(else_result, TerminateSchedulerException):
                    else_result.value = processing_results(
                        data,
                        processing_count,
                        success_count,
                        exception_count,
                        successful_actions,
                        exception_actions,
                    )
                    raise else_result
                result = exception_dict
            else:
                success_count += 1
                successful_actions.append(self.else_action.dict())
                result = else_result
        else:
            success_count += 1
            successful_actions.append(self.test_action.dict())

            if self.if_action:  # execute the if_action
                try:
                    if_result = self.if_action.execute(tag=tag, data=data)
                    self.logger_info(
                        f"IfElse: tag ({tag}); successfully executed 'if' action ({self.if_action})"
                    )
                except Exception as exception:
                    if_result = exception
                    self.logger_exception(
                        f"IfElse: tag ({tag}); error while executing 'if' action ({self.else_action}) exception ({exception})",
                        exc_info=exception,
                    )
                processing_count += 1
                if isinstance(if_result, Exception):
                    exception_count += 1
                    exception_dict = self.if_action.dict().copy()
                    exception_dict.update({"exception": str(if_result)})
                    exception_actions.append(exception_dict)
                    if isinstance(if_result, TerminateSchedulerException):
                        if_result.value = processing_results(
                            data,
                            processing_count,
                            success_count,
                            exception_count,
                            successful_actions,
                            exception_actions,
                        )
                        raise if_result
                    result = exception_dict
                else:
                    success_count += 1
                    successful_actions.append(self.if_action.dict())
                    result = if_result
            else:  # just use test_result since there is no if_action
                result = test_result

        processing_info = processing_results(
            result,
            processing_count,
            success_count,
            exception_count,
            successful_actions,
            exception_actions,
        )
        if (
            processing_info["processing_count"] > 0
            and processing_info["success_count"] == 0
            and self.exception_on_no_success
        ):
            raise Exception(
                "exception generated; all actions failing treated as an if-else action failure",
                json.dumps(self.dict()),
                json.dumps(processing_info),
            )
        extra = processing_info if self.include_processing_info else None
        return self.action_result(
            result=self.get_result(processing_info["result"]), data=data, extra=extra
        )


# experimental


class Compose(Action):
    compose: str = "compose"
    actions: List[Action]

    def description(self):
        return (
            f"Performs function composition on supplied actions in left-to-right order."
        )

    def execute(self, tag: str = None, data: dict = None):
        loop_data = data
        for action in self.actions:
            loop_data = action.execute(tag=tag, data=loop_data)
        return self.action_result(result=self.get_result(loop_data), data=data)



import logging
from pydantic import BaseModel
from datetime import datetime, timedelta
from collections import namedtuple


logger = logging.getLogger(__name__)

"""
ProgramItems are the link between the dispatcher and programs.

type = defer | expire
dt = datetime.datetime
"""
ProgramItem = namedtuple("ProgramItem", ["type", "dt", "scheduler_name", "action_name"])


class Program(BaseModel):
    """
    A Program encapsulates a set of deferrals and expirations for scheduler/action pairs.
    """

    def compute_program_items(self, start: datetime = None, stop: datetime = None):
        """
        Returns a list of ProgramItems used for scheduling by the dispatcher.
        """
        pass



import requests
import logging
import json
from typing import Optional
from whendo.core.action import Action
import whendo.core.util as util_x


logger = logging.getLogger(__name__)


class SendPayload(Action):
    """
    This class sends a payload dictionary to a url.
    """

    url: str
    payload: Optional[dict]

    def description(self):
        return f"This action sends the supplied dictionary payload to ({self.url})."

    def execute(self, tag: str = None, data: dict = None):
        if self.payload:
            payload = self.payload.copy()
            if data:
                payload.update(data)
        elif data:
            payload = data
        else:
            payload = {"result": "no payload"}
        response = requests.post(self.url, payload)
        if response.status_code != requests.codes.ok:
            raise Exception(response)
        return self.action_result(result=response.json(), data=data)


class ExecuteAction(Action):
    """
    Execute an action at host:port.
    """

    host: str
    port: int
    action_name: str

    def description(self):
        return f"This action executes ({self.action_name}) at host:port ({self.host}:{self.port}) unless overriden in the supplied dictionary."

    def execute(self, tag: str = None, data: dict = None):
        self.check_host_port(self.host, self.port)
        if data:
            result = util_x.Http(host=self.host, port=self.port).post_dict(
                f"/actions/{self.action_name}/execute", data
            )
        else:
            result = util_x.Http(host=self.host, port=self.port).get(
                f"/actions/{self.action_name}/execute"
            )
        return self.action_result(result=result, data=data)

    def check_host_port(self, host: str, port: int):
        if host == self.local_host() and port == self.local_port():
            raise ValueError(
                f"supplied host:port ({host}:{port}) must be different from local host:port ({self.local_host()}:{self.local_port()})"
            )



"""
This class implements a single process-based system for defining and scheduling actions local
to the python runtime (api server, testing harness). The class implements all of the potential
endpoints of a restful api and supports testing outside of the restful api implementation.

Instances of this class contain Schedulers, Actions and Programs, which can at any point be submitted to and removed from the
job scheduling mechanism of the schedule library (refer to the 'timed' module).
"""
from pydantic import BaseModel, PrivateAttr
from threading import RLock
from typing import Dict, List
import json
import os
import logging
from datetime import datetime, timedelta
from typing import Optional, Dict, Any
from .util import PP, Dirs, Now, str_to_dt, dt_to_str, DateTime2
from .hooks import DispatcherHooks
from .action import Action
from .program import Program, ProgramItem
from .scheduler import Scheduler, TimedScheduler, ThresholdScheduler, Immediately
from .timed import Timed
from .resolver import resolve_action, resolve_scheduler, resolve_program
from .executor import Executor
from .scheduling import (
    DeferredPrograms,
    DeferredProgram,
    ScheduledActions,
    DatedScheduledActions,
)

logger = logging.getLogger(__name__)


class Dispatcher(BaseModel):
    """
    Serializations of this class are stored in the local file system. When a runtime starts
    up, Dispatch loads the last saved version.
    """

    actions: Dict[str, Action] = {}
    schedulers: Dict[str, Scheduler] = {}
    programs: Dict[str, Program] = {}
    scheduled_actions: ScheduledActions = ScheduledActions()
    deferred_scheduled_actions: DatedScheduledActions = DatedScheduledActions()
    expiring_scheduled_actions: DatedScheduledActions = DatedScheduledActions()
    deferred_programs: DeferredPrograms = DeferredPrograms()
    saved_dir: Optional[str] = None

    # not treated as a model attrs
    _timed: Timed = PrivateAttr(default_factory=Timed.get)
    _timed_for_out_of_band: Timed = PrivateAttr(default_factory=Timed)

    # jobs and timed object
    def set_timed(self, timed: Timed):
        self._timed = timed

    def run_jobs(self):
        return self._timed.run()

    def stop_jobs(self):
        return self._timed.stop()

    def jobs_are_running(self):
        return self._timed.is_running()

    def job_count(self):
        return self._timed.job_count()

    def clear_jobs(self):
        self._timed.clear()

    # internal dispatcher state access
    def get_actions(self):
        with Lok.lock:
            return self.actions

    def get_schedulers(self):
        with Lok.lock:
            return self.schedulers

    def get_programs(self):
        with Lok.lock:
            return self.programs

    def get_scheduled_actions(self):
        with Lok.lock:
            return self.scheduled_actions

    def get_deferred_scheduled_actions(self):
        with Lok.lock:
            return self.deferred_scheduled_actions

    def get_expiring_scheduled_actions(self):
        with Lok.lock:
            return self.expiring_scheduled_actions

    def get_deferred_programs(self):
        return self.deferred_programs

    def get_saved_dir(self):
        return self.saved_dir

    def get_actions_for_scheduler(self, scheduler_name: str):
        with Lok.lock:
            action_names = self.scheduled_actions.actions(scheduler_name)
            return {
                action_name: self.actions[action_name] for action_name in action_names
            }

    def unschedule_scheduler_thunk(self, scheduler_name: str):
        with Lok.lock:
            self.unschedule_scheduler(scheduler_name)

    # other internal dispatcher operations
    def check_for_expirations_and_deferrals(self):
        self.check_for_deferred_programs()
        self.check_for_deferred_actions()
        self.check_for_expiring_actions()

    def initialize(self):
        Lok.reset()
        self._timed_for_out_of_band.clear()
        self._timed_for_out_of_band.every(1).second.do(
            self.check_for_expirations_and_deferrals
        ).tag("check_for_expirations_and_deferrals")
        self._timed_for_out_of_band.run()
        DispatcherHooks.init(
            schedule_program_thunk=lambda program_name, start, stop: self.schedule_program(
                program_name, start, stop
            ),
            unschedule_program_thunk=lambda program_name: self.unschedule_program(
                program_name
            ),
            schedule_action_thunk=lambda scheduler_name, action_name: self.schedule_action(
                scheduler_name, action_name
            ),
            unschedule_scheduler_action_thunk=lambda scheduler_name, action_name: self.unschedule_scheduler_action(
                scheduler_name, action_name
            ),
            unschedule_scheduler_thunk=lambda scheduler_name, action_name: self.unschedule_scheduler(
                scheduler_name
            ),
            defer_action_thunk=lambda scheduler_name, action_name, wait_until: self.defer_action(
                scheduler_name, action_name, wait_until
            ),
            expire_action_thunk=lambda scheduler_name, action_name, expire_on: self.expire_action(
                scheduler_name, action_name, expire_on
            ),
            clear_all_deferred_actions_thunk=lambda scheduler_name, action_name, expire_on: self.clear_all_deferred_actions(),
            clear_all_expiring_actions_thunk=lambda scheduler_name, action_name, expire_on: self.clear_all_expiring_actions(),
        )

    def pprint(self):
        PP.pprint(self.dict())

    def load_current(self):
        return self.load_from_name("current")

    def save_current(self):
        with Lok.lock:
            self.save_to_name("current")

    def load_from_name(self, name: str):
        with Lok.lock:
            if self.saved_dir:
                with open(self.saved_dir + name + ".json", "r") as infile:
                    json_string = json.load(infile)
                    dictionary = json.loads(json_string)
                    return Dispatcher.resolve(dictionary)
            else:
                return None

    def save_to_name(self, name: str):
        with Lok.lock:
            if self.saved_dir:
                with open(self.saved_dir + name + ".json", "w") as outfile:
                    json.dump(self.json(), outfile, indent=2)

    def set_saved_dir(self, saved_dir: str):
        with Lok.lock:
            if saved_dir:
                if not os.path.exists(saved_dir):
                    os.makedirs(saved_dir)
            self.saved_dir = saved_dir
            self.save_current()

    def clear_all(self, should_save: bool = True):
        """
        Removes all actions, schedulers, programs, and foreground
        Timed instance jobs.
        """
        with Lok.lock:
            self.scheduled_actions.clear()
            self.deferred_scheduled_actions.clear()
            self.expiring_scheduled_actions.clear()
            self.deferred_programs.clear()
            self.actions.clear()
            self.schedulers.clear()
            self.programs.clear()
            self._timed.clear()
            if should_save:
                self.save_current()

    def replace_all(self, replacement: object):
        """
        Note #1: the assumption is that the replacement is a Dispatcher. There's
        an module import circularity somewhere. Have yet to find a solution beyond
        this one. This will suffice for now. The submethod is there to show
        intention.

        1. except for the saved_dir, clear everything first
        2. replace elements with replacement elements
        3. save

        Note #2: after [1] all previous actions and schedules are no longer in existence.
        If processing should continue, invoke reschedule_all_schedulers() after
        invoking this method.
        """

        def typed_replace_all(replacement: Dispatcher):
            with Lok.lock:
                self.clear_all(should_save=False)
                self.actions = replacement.get_actions()
                self.schedulers = replacement.get_schedulers()
                self.scheduled_actions = replacement.get_scheduled_actions()
                self.deferred_scheduled_actions = (
                    replacement.get_deferred_scheduled_actions()
                )
                self.expiring_scheduled_actions = (
                    replacement.get_expiring_scheduled_actions()
                )
                self.deferred_programs = replacement.get_deferred_programs()
                self.save_current()

        typed_replace_all(replacement)

    def describe_all(self):
        """
        Returns descriptions of all actions, schedulers and programs.
        """
        result = {}

        stuff = self.actions.copy()
        for name in stuff:
            stuff[name] = stuff[name].description()
        result["actions"] = stuff

        stuff = self.schedulers.copy()
        for name in stuff:
            stuff[name] = stuff[name].description()
        result["schedulers"] = stuff

        stuff = self.programs.copy()
        for name in stuff:
            stuff[name] = stuff[name].description()
        result["programs"] = stuff

        return result

    # actions
    def get_action(self, action_name: str):
        with Lok.lock:
            return self.actions.get(action_name, None)

    def describe_action(self, action_name: str):
        with Lok.lock:
            action = self.get_action(action_name)
            return (
                action.description()
                if action
                else f"action ({action_name}) does not exist."
            )

    def add_action(self, action_name: str, action: Action):
        with Lok.lock:
            self.check_action_name(action_name, invert=True)
            self.actions[action_name] = action
            self.save_current()

    def set_action(self, action_name: str, action: Action):
        with Lok.lock:
            self.check_action_name(action_name)
            self.actions[action_name] = action
            self.save_current()

    def delete_action(self, action_name: str):
        with Lok.lock:
            self.check_action_name(action_name)
            deleted_schedulers = self.scheduled_actions.delete_action(action_name)
            for ds in deleted_schedulers:
                self.check_scheduler(ds)
            self.actions.pop(action_name, None)
            self.deferred_scheduled_actions.delete_dated_action(action_name=action_name)
            self.expiring_scheduled_actions.delete_dated_action(action_name=action_name)

            # delete programs referencing action_name
            programs_copy = self.programs.copy()
            for program_name in programs_copy:
                program_items = programs_copy[program_name].compute_program_items()
                for program_item in program_items:
                    if program_item.action_name == action_name:
                        self.delete_program(program_name)
                        break

            self.save_current()

    def execute_action(self, action_name: str):
        with Lok.lock:
            self.check_action_name(action_name)
            return self.get_action(action_name).execute()

    def execute_action_with_data(self, action_name: str, data: dict):
        with Lok.lock:
            self.check_action_name(action_name)
            return self.get_action(action_name).execute(data=data)

    def execute_supplied_action(self, supplied_action: Action):
        with Lok.lock:
            return supplied_action.execute()

    # schedulers
    def get_scheduler(self, scheduler_name: str):
        with Lok.lock:
            return self.schedulers.get(scheduler_name, None)

    def describe_scheduler(self, scheduler_name: str):
        with Lok.lock:
            scheduler = self.get_scheduler(scheduler_name)
            return (
                scheduler.description()
                if scheduler
                else f"scheduler ({scheduler_name}) does not exist."
            )

    def add_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        with Lok.lock:
            self.check_scheduler_name(scheduler_name, invert=True)
            self.schedulers[scheduler_name] = scheduler
            self.save_current()

    def set_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            self.schedulers[scheduler_name] = scheduler
            self.reschedule_scheduler(scheduler_name)
            self.save_current()

    def delete_scheduler(self, scheduler_name: str):
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            self.unschedule_scheduler(scheduler_name)
            self.schedulers.pop(scheduler_name)
            self.scheduled_actions.delete_scheduler(scheduler_name)
            self.deferred_scheduled_actions.delete_dated_scheduler(
                scheduler_name=scheduler_name
            )
            self.expiring_scheduled_actions.delete_dated_scheduler(
                scheduler_name=scheduler_name
            )

            # delete programs referencing scheduler_name
            programs_copy = self.programs.copy()
            for program_name in programs_copy:
                program_items = programs_copy[program_name].compute_program_items()
                for program_item in program_items:
                    if program_item.scheduler_name == scheduler_name:
                        self.delete_program(program_name)
                        break

            self.save_current()

    def check_scheduler(self, scheduler_name: str):
        self.check_scheduler_name(scheduler_name)
        if len(self.scheduled_actions.actions(scheduler_name)) == 0:
            scheduler = self.get_scheduler(scheduler_name)
            if isinstance(scheduler, TimedScheduler):
                scheduler.set_timed(self._timed)
            scheduler.unschedule(scheduler_name)

    # programs
    def get_program(self, program_name: str):
        with Lok.lock:
            return self.programs.get(program_name, None)

    def describe_program(self, program_name: str):
        with Lok.lock:
            program = self.get_program(program_name)
            return (
                program.description()
                if program
                else f"program ({program_name}) does not exist."
            )

    def add_program(self, program_name: str, program: Program):
        with Lok.lock:
            self.check_program_name(program_name, invert=True)
            self.check_program(program)
            self.programs[program_name] = program
            self.save_current()

    def set_program(self, program_name: str, program: Program):
        with Lok.lock:
            self.check_program_name(program_name)
            self.check_program(program)
            self.programs[program_name] = program
            self.save_current()

    def delete_program(self, program_name: str):
        """
        Deletes program from programs and removes all references
        in deferred_scheduled_programs.
        """
        with Lok.lock:
            self.check_program_name(program_name)
            self.unschedule_program(program_name)
            self.programs.pop(program_name)
            self.save_current()

    def unschedule_program(self, program_name: str):
        """
        Deletes program from programs and removes all references
        in deferred_scheduled_programs.
        """
        with Lok.lock:
            self.check_program_name(program_name)
            self.deferred_programs.clear_program(program_name)
            self.save_current()

    def check_program(self, program: Program):
        """
        Makes sure that actions and schedulers referenced in the program exist.
        """
        with Lok.lock:
            program_items = program.compute_program_items()
            error_msgs = []
            if len(program_items) == 0:
                error_msgs.append(f"empty program")
            else:
                for item in program_items:
                    if item.action_name not in self.actions:
                        error_msgs.append(f"missing action: ({item.action_name})")
                    if item.scheduler_name not in self.schedulers:
                        error_msgs.append(f"missing scheduler: ({item.scheduler_name})")
            if len(error_msgs) > 0:
                raise ValueError(", ".join(error_msgs))

    def schedule_program(self, program_name: str, start: datetime, stop: datetime):
        """
        This method defers the scheduling of a program.
        """
        with Lok.lock:
            self.check_program_name(program_name)
            program = self.programs[program_name]
            self.check_program(program)
            self.deferred_programs.add(DeferredProgram(program_name, start, stop))
            self.save_current()

    def check_for_deferred_programs(self):
        """
        This gets run as a job in the out-of-band Timed instance.

        It looks for deferred programs whose start dates are prior to the current
        time and 'disseminates' these programs. See the
        schedule_program and initialize methods for more details.
        """
        with Lok.lock:
            for dp in self.deferred_programs.pop():
                try:
                    self.disseminate_program(dp.program_name, dp.start, dp.stop)
                except Exception as exception:
                    logger.error(
                        f"failed to dissemminate program ({program_name}) using start ({start}), ({stop})",
                        exception,
                    )
                self.save_current()

    def disseminate_program(self, program_name: str, start: datetime, stop: datetime):
        """
        Defers and expires scheduler/actions as necessary. Once disseminated,
        the associated scheduler/actions are dissassociated from the originating
        program. The chickens have flown the coop. They are now scheduled and running
        independently.
        """
        with Lok.lock:
            self.check_program_name(program_name)
            program = self.programs[program_name]
            self.check_program(program)
            program_items = program.compute_program_items(start, stop)
            for item in program_items:
                if item.type == "defer":
                    self.defer_action(item.scheduler_name, item.action_name, item.dt)
                elif item.type == "expire":
                    self.expire_action(item.scheduler_name, item.action_name, item.dt)

    def get_deferred_program_count(self):
        # returns the total number of datetime2s in the deferred programs dictionary (a dictionary
        # of dictionaries)
        with Lok.lock:
            return self.deferred_programs.count()

    def clear_all_deferred_programs(self):
        with Lok.lock:
            self.deferred_programs.clear()
            self.save_current()

    # scheduling
    def schedule_action(self, scheduler_name: str, action_name: str):
        """
        Puts the scheduler/action into active processing. The scheduled_action
        dictionary mirrors the successful scheduling of actions.
        """
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            self.check_action_name(action_name)
            scheduler = self.get_scheduler(scheduler_name)
            if isinstance(scheduler, TimedScheduler):
                scheduler.set_timed(self._timed)
            elif isinstance(
                scheduler, Immediately
            ):  # executes once; does not participate in scheduling
                tag = f"{scheduler_name}:{action_name}"
                action = self.get_action(action_name)
                try:
                    action.execute(tag=tag)
                    logger.info(f"Execution: tag ({tag}); executed action ({action})")
                except Exception as exception:
                    logger.exception(
                        f"Execution: tag ({tag}); error while executing action ({action})",
                        exc_info=exception,
                    )
                return
            action_names = self.scheduled_actions.actions(scheduler_name)
            if action_name in action_names:
                return  # don't need to schedule
            else:
                self.scheduled_actions.add(scheduler_name, action_name)
                action_names = self.scheduled_actions.actions(scheduler_name)
                if len(action_names) == 1:  # > 1 implies already scheduled
                    scheduler.schedule(
                        scheduler_name,
                        Executor(
                            self.get_actions_for_scheduler,
                            self.unschedule_scheduler_thunk,
                        ),
                    )
            self.save_current()

    def unschedule_scheduler_action(self, scheduler_name: str, action_name: str):
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            self.check_action_name(action_name)
            # get the scheduler
            scheduler = self.get_scheduler(scheduler_name)
            if isinstance(scheduler, TimedScheduler):
                scheduler.set_timed(self._timed)
            # remove the action from the scheduler's actions
            scheduler_name = self.scheduled_actions.delete(scheduler_name, action_name)
            # unschedule the scheduler if a non-None scheduler name is returned
            if scheduler_name:
                scheduler.unschedule(scheduler_name)
            self.save_current()

    def unschedule_scheduler(self, scheduler_name: str):
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            scheduler = self.get_scheduler(scheduler_name)
            if isinstance(scheduler, TimedScheduler):
                scheduler.set_timed(self._timed)
            scheduler.unschedule(scheduler_name)
            self.scheduled_actions.delete_scheduler(scheduler_name)
            self.save_current()

    def reschedule_scheduler(self, scheduler_name: str):
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            scheduler = self.get_scheduler(scheduler_name)
            if isinstance(scheduler, TimedScheduler):
                scheduler.set_timed(self._timed)
            scheduler.schedule(
                scheduler_name,
                Executor(
                    self.get_actions_for_scheduler, self.unschedule_scheduler_thunk
                ),
            )

    def reschedule_all_schedulers(self):
        with Lok.lock:
            for scheduler_name in self.scheduled_actions.scheduler_names():
                self.reschedule_scheduler(scheduler_name)

    def get_scheduled_action_count(self):
        # returns the total number of actions in the scheduled_actions dictionary
        # should be equal to the number of jobs in related Timed instance
        with Lok.lock:
            return self.scheduled_actions.action_count()

    # defer scheduler/action
    def defer_action(self, scheduler_name: str, action_name: str, wait_until: datetime):
        """
        This method defers the start of scheduling an action. The data structure is a dictionary
        with a datetime as the key and a scheduled_actions style dictionary as the value.
        """
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            self.check_action_name(action_name)
            self.deferred_scheduled_actions.apply_date(
                scheduler_name=scheduler_name,
                action_name=action_name,
                date_time=wait_until,
            )
            self.save_current()

    def check_for_deferred_actions(self):
        """
        This gets run as a job in the out-of-band Timed instance.

        It looks for deferred actions whose dates are prior to the current
        time and schedules them using the associated scheduler. See the
        defer_action and initialize methods for more details.
        """
        with Lok.lock:
            self.deferred_scheduled_actions.check_for_dated_actions(
                schedule_update_thunk=self.schedule_action,
                verb="schedule",
            )
            self.save_current()

    def get_deferred_action_count(self):
        # returns the total number of actions in the deferred actions dictionary (a dictionary
        # of dictionaries)
        with Lok.lock:
            return self.deferred_scheduled_actions.action_count()

    def clear_all_deferred_actions(self):
        with Lok.lock:
            self.deferred_scheduled_actions.clear()
            self.save_current()

    # expire scheduler/action
    def expire_action(self, scheduler_name: str, action_name: str, expire_on: datetime):
        """
        This method expires an action with scheduler. The data structure is a dictionary
        with a datetime as the key and a scheduled_actions style dictionary as the value.
        """
        with Lok.lock:
            self.check_scheduler_name(scheduler_name)
            self.check_action_name(action_name)
            self.expiring_scheduled_actions.apply_date(
                scheduler_name=scheduler_name,
                action_name=action_name,
                date_time=expire_on,
            )

            self.save_current()

    def check_for_expiring_actions(self):
        """
        This gets run as a job in the out-of-band Timed instance.

        It looks for expiring actions whose dates are later than the current
        time and unschedules them within the context of the associated scheduler.
        See the expire_action and initialize methods for more details.
        """
        with Lok.lock:
            self.expiring_scheduled_actions.check_for_dated_actions(
                schedule_update_thunk=self.unschedule_scheduler_action,
                verb="unschedule",
            )
            self.save_current()

    def get_expiring_action_count(self):
        # returns the total number of actions in the deferred actions dictionary (a dictionary
        # of dictionaries)
        with Lok.lock:
            return self.expiring_scheduled_actions.action_count()

    def clear_all_expiring_actions(self):
        with Lok.lock:
            self.expiring_scheduled_actions.clear()
            self.save_current()

    # utility
    def check_scheduler_name(self, scheduler_name: str, invert: bool = False):
        if invert:
            assert (
                scheduler_name not in self.schedulers
            ), f"scheduler ({scheduler_name}) already exists"
        else:
            assert (
                scheduler_name in self.schedulers
            ), f"scheduler ({scheduler_name}) does not exist"

    def check_action_name(self, action_name: str, invert: bool = False):
        if invert:
            assert (
                action_name not in self.actions
            ), f"action ({action_name}) already exists"
        else:
            assert action_name in self.actions, f"action ({action_name}) does not exist"

    def check_program_name(self, program_name: str, invert: bool = False):
        if invert:
            assert (
                program_name not in self.programs
            ), f"program ({program_name}) already exists"
        else:
            assert (
                program_name in self.programs
            ), f"program ({program_name}) does not exist"

    @classmethod
    def resolve(cls, dictionary: dict = {}):
        """
        converts dictionary to an Dispatcher instance.
        """
        if len(dictionary) == 0:
            return Dispatcher()
        else:
            saved_dir = dictionary["saved_dir"]
            actions = dictionary["actions"]
            schedulers = dictionary["schedulers"]
            programs = dictionary["programs"]
            scheduled_actions = dictionary["scheduled_actions"]
            deferred_scheduled_actions = dictionary["deferred_scheduled_actions"]
            expiring_scheduled_actions = dictionary["expiring_scheduled_actions"]
            deferred_programs = dictionary["deferred_programs"]
            # deferred_scheduled_programs = dictionary["deferred_scheduled_programs"]
            # replace key's value for each key...
            for action_name in actions:
                actions[action_name] = resolve_action(actions[action_name])
            for scheduler_name in schedulers:
                schedulers[scheduler_name] = resolve_scheduler(
                    schedulers[scheduler_name]
                )
            for program_name in programs:
                programs[program_name] = resolve_program(programs[program_name])
            return Dispatcher(
                saved_dir=saved_dir,
                actions=actions,
                schedulers=schedulers,
                programs=programs,
                scheduled_actions=scheduled_actions,
                deferred_scheduled_actions=deferred_scheduled_actions,
                expiring_scheduled_actions=expiring_scheduled_actions,
                deferred_programs=deferred_programs,
            )


class Lok:
    """
    usage:
        with Lok.lock:
            # critical section
    note:
        singleton
    """

    lock = RLock()

    @classmethod
    def reset(cls):
        cls.lock = RLock()


class DispatcherSingleton:
    # this call returns the standard non-test Data singleton
    dispatcher = None

    @classmethod
    def get(cls):
        if not cls.dispatcher:
            cls.dispatcher = Dispatcher(saved_dir=Dirs.saved_dir())
            try:
                cls.dispatcher = cls.dispatcher.load_current()
            except Exception as exception:
                logger.error("error loading dispatcher from disk", exception)
            cls.dispatcher.set_timed(Timed.get())
            cls.dispatcher.initialize()
            cls.dispatcher.reschedule_all_schedulers()
        return cls.dispatcher
